﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Martian_WinForms
{
    public abstract class Movable
    {
        public static List<Movable> AllEntities { get; set; } = new List<Movable>();
        protected Movable(int x, int y, Bitmap spriteImage)
        {
            X = x;
            Y = y;
            NextX = x;
            NextY = y;
            SpriteImage = spriteImage;
            AllEntities.Add(this);
        }

        public int X { get; set; }
        public int Y { get; set; }
        public Point Position
        {
            get
            {
                return new Point(X, Y);
            }
        }
        public Bitmap SpriteImage { get; set; }
        // Точка к которой движется персонаж.
        public int NextX { get; set; }
        public int NextY { get; set; }
        // Отклонение персонажа от центра текущего тайла.
        public float OffsetX { get; set; }
        public  float OffsetY { get; set; }
        // Точные координаты персонажа с учётом отклонения от центра.
        public float GetActualX
        {
            get
            {
                return X * Settings.BlockSize + OffsetX;
            }
        }
        public float GetActualY
        {
            get
            {
                return Y * Settings.BlockSize + OffsetY;
            }
        }
        // Скорость, с которой игрок перемещается между тайлами.
        public float Speed { get; set; } = 4f;

        // Является ли текущий шаг чётным по счёту с начала игры.
        public bool IsEvenStep { get; set; } = true;
        public bool IsMoving
        {
            get
            {
                return X != NextX || Y != NextY;
            }
        }

        // Определяет, какую часть спрайта нужно отрисовывать в данный момент
        public int XSpriteOffset { get; set; }

        /// <summary>
        /// Метод вызывается каждый игровой кадр. Персонаж смещается по направлению к тайлу, к которому движется, имитируя движение.
        /// </summary>
        public virtual void MoveTowardsNext()
        {
            if (!IsMoving) return;
            int spriteOffset = 0;
            if (NextX > X)
            {
                OffsetX += Speed;
                spriteOffset = 32;
                if (OffsetX > Settings.BlockSize)
                    HandleReachedTileCenter();
            }
            else if (NextX < X)
            {
                OffsetX -= Speed;
                spriteOffset = 96;
                if (OffsetX < -Settings.BlockSize)
                    HandleReachedTileCenter();
            }
            if (NextY > Y)
            {
                OffsetY += Speed;
                spriteOffset = 0;
                if (OffsetY > Settings.BlockSize)
                    HandleReachedTileCenter();
            }
            else if (NextY < Y)
            {
                OffsetY -= Speed;
                spriteOffset = 64;
                if (OffsetY < -Settings.BlockSize)
                    HandleReachedTileCenter();
            }
            if (Math.Abs(OffsetX) > Settings.BlockSize / 2 || Math.Abs(OffsetY) > Settings.BlockSize / 2)
                spriteOffset += 128;
            if (IsEvenStep)
                spriteOffset += 128;
            if (this as Cube == null)
                XSpriteOffset = spriteOffset;
        }

        public virtual void Move(Direction direction)
        {
            switch (direction)
            {
                case Direction.Up: NextY--; break;
                case Direction.Right: NextX++; break;
                case Direction.Down: NextY++; break;
                case Direction.Left: NextX--; break;
                case Direction.None: return;
            }
            // Вернуть следующие координаты к прежнему состоянию,
            // если на них стена или закрытая дверь.
            if (Common.CurrentMap.IsWall(NextX, NextY) || Common.CurrentMap.IsLockedDoor(NextX, NextY))
            {
                NextX = X;
                NextY = Y;
                return;
            }
            IsEvenStep = !IsEvenStep;
        }

        /// <summary>
        /// Вызывается, когда персонаж добирается до середины нового тайла, его координаты и смещение от центра(Offset) обновляются, применяются эффекты тайла, подбираются ключ-карты
        /// </summary>
        public virtual void HandleReachedTileCenter()
        {
            if (Settings.ConsoleMode) ConsoleMode.RedrawTile(X, Y, false);
            OffsetX = 0;
            X = NextX;
            OffsetY = 0;
            Y = NextY;
        }
        public Direction GetDirection(Point pointTo)
        {
            if (X > pointTo.X)
            {
                return Direction.Left;
            }
            else if (X < pointTo.X)
            {
                return Direction.Right;
            }
            if (Y > pointTo.Y)
            {
                return Direction.Up;
            }
            else if (Y < pointTo.Y)
            {
                return Direction.Down;
            }
            return Direction.None;
        }

        public Direction CurrentDirection { get; set; }

        public abstract void Act();
    }
}
